package java.awt.geom;

import sun.awt.geom.Crossings;
import java.awt.Rectangle;
import java.util.Enumeration;
import sun.awt.geom.AreaOp;
import sun.awt.geom.Curve;
import java.util.Vector;
import java.awt.Shape;

public class Area implements Shape, Cloneable
{
    private static Vector EmptyCurves;
    private Vector curves;
    private Rectangle2D cachedBounds;
    
    public Area() {
        this.curves = Area.EmptyCurves;
    }
    
    public Area(Shape shape) {
        if ((shape instanceof Area)) {
            this.curves = ((Area)shape).curves;
        }
        else {
            this.curves = pathToCurves(shape.getPathIterator(null));
        }
    }
    
    private static Vector pathToCurves(PathIterator pathIterator) {
        Vector<Curve> vector = new Vector();
        int windingRule = pathIterator.getWindingRule();
        double[] array = new double[23];
        double n = 0.0;
        double n2 = 0.0;
        double n3 = 0.0;
        double n4 = 0.0;
        while (!pathIterator.isDone()) {
            switch (pathIterator.currentSegment(array)) {
                case 0: {
                    Curve.insertLine(vector, n3, n4, n, n2);
                    n = (n3 = array[0]);
                    n2 = (n4 = array[1]);
                    Curve.insertMove(vector, n, n2);
                    break;
                }
                case 1: {
                    double x1 = array[0];
                    double y1 = array[1];
                    Curve.insertLine(vector, n3, n4, x1, y1);
                    n3 = x1;
                    n4 = y1;
                    break;
                }
                case 2: {
                    double n5 = array[2];
                    double n6 = array[3];
                    Curve.insertQuad(vector, n3, n4, array);
                    n3 = n5;
                    n4 = n6;
                    break;
                }
                case 3: {
                    double n7 = array[4];
                    double n8 = array[5];
                    Curve.insertCubic(vector, n3, n4, array);
                    n3 = n7;
                    n4 = n8;
                    break;
                }
                case 4: {
                    Curve.insertLine(vector, n3, n4, n, n2);
                    n3 = n;
                    n4 = n2;
                    break;
                }
            }
            pathIterator.next();
        }
        Curve.insertLine(vector, n3, n4, n, n2);
        Object o;
        if (windingRule == 0) {
            o = new AreaOp.EOWindOp();
        }
        else {
            o = new AreaOp.NZWindOp();
        }
        return o.calculate(vector, Area.EmptyCurves);
    }
    
    public void add(Area area) {
        this.curves = new AreaOp.AddOp().calculate(this.curves, area.curves);
        this.invalidateBounds();
    }
    
    public void subtract(Area area) {
        this.curves = new AreaOp.SubOp().calculate(this.curves, area.curves);
        this.invalidateBounds();
    }
    
    public void intersect(Area area) {
        this.curves = new AreaOp.IntOp().calculate(this.curves, area.curves);
        this.invalidateBounds();
    }
    
    public void exclusiveOr(Area area) {
        this.curves = new AreaOp.XorOp().calculate(this.curves, area.curves);
        this.invalidateBounds();
    }
    
    public void reset() {
        this.curves = new Vector();
        this.invalidateBounds();
    }
    
    public boolean isEmpty() {
        return this.curves.size() == 0;
    }
    
    public boolean isPolygonal() {
        Enumeration<Object> elements = this.curves.elements();
        while (elements.hasMoreElements()) {
            if (((Curve)elements.nextElement()).getOrder() <= 1) {
                continue;
            }
            return false;
        }
        return true;
    }
    
    public boolean isRectangular() {
        int size = this.curves.size();
        if (size == 0) {
            return true;
        }
        if (size > 3) {
            return false;
        }
        Curve curve = (Curve)this.curves.get(1);
        Curve curve2 = (Curve)this.curves.get(2);
        return curve.getOrder() == 1 && curve2.getOrder() == 1 && curve.getXTop() == curve.getXBot() && curve2.getXTop() == curve2.getXBot() && curve.getYTop() == curve2.getYTop() && curve.getYBot() == curve2.getYBot();
    }
    
    public boolean isSingular() {
        if (this.curves.size() < 3) {
            return true;
        }
        Enumeration<Object> elements = this.curves.elements();
        elements.nextElement();
        while (elements.hasMoreElements()) {
            if (((Curve)elements.nextElement()).getOrder() != 0) {
                continue;
            }
            return false;
        }
        return true;
    }
    
    private void invalidateBounds() {
        this.cachedBounds = null;
    }
    
    private Rectangle2D getCachedBounds() {
        if (this.cachedBounds != null) {
            return this.cachedBounds;
        }
        Rectangle2D.Double cachedBounds = new Rectangle2D.Double();
        if (this.curves.size() > 0) {
            Curve curve = (Curve)this.curves.get(0);
            cachedBounds.setRect(curve.getX0(), curve.getY0(), 0.0, 0.0);
            for (int i = 1; i < this.curves.size(); ++i) {
                ((Curve)this.curves.get(i)).enlarge(cachedBounds);
            }
        }
        return this.cachedBounds = cachedBounds;
    }
    
    @Override
    public Rectangle2D getBounds2D() {
        return this.getCachedBounds().getBounds2D();
    }
    
    @Override
    public Rectangle getBounds() {
        return this.getCachedBounds().getBounds();
    }
    
    public Object clone() {
        return new Area(this);
    }
    
    public boolean equals(Area area) {
        return area == this || (area != null && new AreaOp.XorOp().calculate(this.curves, area.curves).isEmpty());
    }
    
    public void transform(AffineTransform affineTransform) {
        if (affineTransform == null) {
            throw new NullPointerException("transform must not be null");
        }
        this.curves = pathToCurves(this.getPathIterator(affineTransform));
        this.invalidateBounds();
        return;
    }
    
    public Area createTransformedArea(AffineTransform affineTransform) {
        Area area = new Area(this);
        area.transform(affineTransform);
        return area;
    }
    
    @Override
    public boolean contains(double x, double y) {
        if (!this.getCachedBounds().contains(x, y)) {
            return false;
        }
        Enumeration<Object> elements = this.curves.elements();
        int n = 0;
        while (elements.hasMoreElements()) {
            n += ((Curve)elements.nextElement()).crossingsFor(x, y);
        }
        return (n & 0x1) == 0x1;
    }
    
    @Override
    public boolean contains(Point2D point2D) {
        return this.contains(point2D.getX(), point2D.getY());
    }
    
    @Override
    public boolean contains(double xlo, double ylo, double n, double n2) {
        if (n < 0.0 || n2 < 0.0) {
            return false;
        }
        if (!this.getCachedBounds().contains(xlo, ylo, n, n2)) {
            return false;
        }
        Crossings crossings = Crossings.findCrossings(this.curves, xlo, ylo, xlo + n, ylo + n2);
        return crossings != null && crossings.covers(ylo, ylo + n2);
    }
    
    @Override
    public boolean contains(Rectangle2D rectangle2D) {
        return this.contains(rectangle2D.getX(), rectangle2D.getY(), rectangle2D.getWidth(), rectangle2D.getHeight());
    }
    
    @Override
    public boolean intersects(double xlo, double ylo, double n, double n2) {
        if (n < 0.0 || n2 < 0.0) {
            return false;
        }
        if (!this.getCachedBounds().intersects(xlo, ylo, n, n2)) {
            return false;
        }
        Crossings crossings = Crossings.findCrossings(this.curves, xlo, ylo, xlo + n, ylo + n2);
        return crossings == null || !crossings.isEmpty();
    }
    
    @Override
    public boolean intersects(Rectangle2D rectangle2D) {
        return this.intersects(rectangle2D.getX(), rectangle2D.getY(), rectangle2D.getWidth(), rectangle2D.getHeight());
    }
    
    @Override
    public PathIterator getPathIterator(AffineTransform affineTransform) {
        return new AreaIterator(this.curves, affineTransform);
    }
    
    @Override
    public PathIterator getPathIterator(AffineTransform affineTransform, double n) {
        return new FlatteningPathIterator(this.getPathIterator(affineTransform), n);
    }
    
    static {
        Area.EmptyCurves = new Vector();
    }
}
